#include <linux/stop_machine.h>
#include <linux/kallsyms.h>
#include <linux/preempt.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/blkdev.h>
#include <linux/version.h>
#include "kernel_hook.h"
#include "lbb_kernel.h"
#include "lbb_device.h"
#include "lbb_mapper.h"
#include "info.h"
#define HOOK_FULL_LEN (16)
#if LINUX_VERSION_CODE >=KERNEL_VERSION(6,1,0)
#define HOOK_HEX {0x00,0x00,0x00,0x00,0x90,0x55,0x48,0x89,0xe5,0x41,0x55,0x41,0x54,0x49,0x89,0xfc}
#define HOOK_SET (5)
#elif LINUX_VERSION_CODE >=KERNEL_VERSION(4,5,0)
#define HOOK_HEX {0x0f,0x1f,0x44,0x00,0x00,0x55,0x48,0x89,0xe5,0x41,0x54,0x53,0x48,0x89,0xfb,0x48}
#define HOOK_SET (5)
#elif LINUX_VERSION_CODE >=KERNEL_VERSION(3,10,0)
#define HOOK_HEX {0x0f,0x1f,0x44,0x00,0x00,0x55,0x48,0x89,0xe5,0x41,0x55,0x41,0x54,0x53,0x48,0x89}
#define HOOK_SET (5)
#elif LINUX_VERSION_CODE >=KERNEL_VERSION(2,6,30)
#define HOOK_HEX {0x55,0x48,0x89,0xe5,0x41,0x54,0x53,0x48,0x83,0xec,0x30,0x0f,0x1f,0x44,0x00,0x00,}
#define HOOK_SET (6)
#elif LINUX_VERSION_CODE >=KERNEL_VERSION(2,6,0)
#define HOOK_HEX {0x55,0x48,0x89,0xe5,0x48,0x83,0xec,0x40,0x48,0x89,0x5d,0xf0,0x4c,0x89,0x65,0xf8}
#define HOOK_SET (8)
#endif
#define HOOK_USED_LEN (HOOK_SET+1)
typedef struct
{
atomic_t usage;
unsigned char offset;
unsigned char target[HOOK_FULL_LEN];
void * fn_hook;
void * fn_store;
void * map_target;
void * fn_orig;
void * map_orig;
} knook;
void notrace submit_bio_orig(void)
{
asm volatile (".rept 0x20\n" ".byte 0x90\n" ".endr\n");
}
extern knook k_submit_bio;
static void * map_writable(void * addr,size_t len)
{
void *vaddr;
int nr_pages=DIV_ROUND_UP(offset_in_page(addr)+len,PAGE_SIZE);
struct page **pages=xkmalloc(nr_pages * sizeof(*pages));
void *page_addr=(void *)((unsigned long)addr & PAGE_MASK);
int i;
if (pages==NULL) return NULL;
for (i=0;i < nr_pages;i++)
{
if (__module_address((unsigned long)page_addr)==NULL)
{
pages[i]=virt_to_page(page_addr);
WARN_ON(!PageReserved(pages[i]));
}
else
{
pages[i]=vmalloc_to_page(page_addr);
}
if (pages[i]==NULL)
{
xkfree(pages);
return NULL;
}
page_addr+=PAGE_SIZE;
}
vaddr=vmap(pages,nr_pages,VM_MAP,PAGE_KERNEL);
xkfree(pages);
if (vaddr==NULL) return NULL;
return vaddr+offset_in_page(addr);
}
static inline void x86_put_jmp( void * a,void * f,void * t)
{
*((char *)(a+0))=0xE9;
*(( int *)(a+1))=(long)(t-(f+5));
}
static int do_init_one_hook(void * arg)
{
knook * kn=(knook *)arg;
if (kn && atomic_read(&kn->usage)==1)
{
x86_put_jmp((unsigned char *)kn->map_target,(unsigned char *)kn->fn_store,(unsigned char *)kn->fn_hook);
}
return 0;
}
static int init_one_hook(knook * kn)
{
int i;
unsigned char * fp=(unsigned char *)kn->fn_store;
atomic_set(&kn->usage,0);
kn->map_target=NULL;
kn->map_orig=NULL;
for (i=0;i < HOOK_USED_LEN;++i)
if (kn->target[i] !=fp[i]) return-1;
kn->map_orig=map_writable(kn->fn_orig,HOOK_FULL_LEN);
if (kn->map_orig==NULL) return-2;
kn->map_target=map_writable(kn->fn_store,HOOK_FULL_LEN);
if (kn->map_target==NULL)
{
vunmap((void *)((unsigned long)kn->map_orig & PAGE_MASK));
kn->map_orig=NULL;
return-3;
}
memcpy(kn->map_orig,kn->target,HOOK_FULL_LEN);
x86_put_jmp(kn->map_orig+kn->offset,kn->fn_orig+kn->offset,kn->fn_store+kn->offset);
atomic_inc(&kn->usage);
stop_machine(do_init_one_hook,kn,NULL);
PrintLog(0,KERN_INFO "Initialize kernel hooks success. (Kernel %s)\n",print_version(LINUX_VERSION_CODE));
return 0;
}
int init_hooks(void)
{
knook * kn;
int rcod=0;
k_submit_bio.fn_store=submit_bio;
kn=&k_submit_bio;
rcod=init_one_hook(kn);
if (rcod)
{
if (rcod==-1)
{
PrintLog(0,KERN_ERR "Error create kernel hook. Different with template.\n");
print_vect(&(kn->target),HOOK_FULL_LEN,"\n");
print_vect(kn->fn_store,HOOK_FULL_LEN,"\n");
PrintLog(0,KERN_NOTICE "Kernel %s\n",print_version(LINUX_VERSION_CODE));
}
else PrintLog(0,KERN_ERR "Error create kernel hook (%d)\n",rcod);
return-999;
}
return 0;
}
static int do_cleanup_one_hook(void * arg)
{
knook * kn=(knook *)arg;
if (atomic_read(&kn->usage))
memcpy(kn->map_target,kn->target,HOOK_USED_LEN);
return 0;
}
static int cleanup_one_hook(knook * kn)
{
int i=0;
stop_machine(do_cleanup_one_hook,kn,NULL);
while (atomic_read(&kn->usage) !=1)
{
if (++i > 20) return-1;
msleep_interruptible(100);
}
if (kn->map_target) vunmap((void *)((unsigned long)kn->map_target & PAGE_MASK));
if (kn->map_orig) vunmap((void *)((unsigned long)kn->map_orig & PAGE_MASK));
atomic_set(&kn->usage,0);
kn->map_target=NULL;
kn->map_orig=NULL;
return 0;
}
void cleanup_hooks(void)
{
int rcod=cleanup_one_hook(&k_submit_bio);
if (rcod) PrintLog(0,KERN_ERR "Error remove kernel hook (%d)\n",rcod);
PrintLog(0,KERN_INFO "Cleanup all kernel hooks success.\n");
}
#if LINUX_VERSION_CODE >=KERNEL_VERSION(4,5,0)
typedef blk_qc_t (* fn_submit_bio)(struct bio *);
blk_qc_t submit_bio_knook(struct bio * bio)
{
int rcod=0;
blk_qc_t result=BLK_QC_T_NONE;
struct lbb_device * xdev;
fn_submit_bio fn_orig=(fn_submit_bio)submit_bio_orig;
atomic_inc(&k_submit_bio.usage);
if (bio_data_dir(bio)==WRITE)
{
#if (INFO_BIO_TYPE==2)
xdev=lbb_find_table(bdget_disk(bio->bi_disk,bio->bi_partno));
#elif (INFO_BIO_TYPE==1)
xdev=lbb_find_table(bio->bi_bdev);
#else
#error Undefined struct bio type.
#endif
if (xdev !=NULL)
{
rcod=lbb_check_block_bio_changed(xdev,bio);
}
}
if (rcod==0) fn_orig(bio);
atomic_dec(&k_submit_bio.usage);
return result;
}
blk_qc_t khook_submit_bio_origin(struct bio * bio,unsigned int opf)
{
blk_qc_t result;
fn_submit_bio fn_orig=(fn_submit_bio)submit_bio_orig;
if (opf) bio_set_op_attrs(bio,opf,0);
atomic_inc(&k_submit_bio.usage);
result=fn_orig(bio);
atomic_dec(&k_submit_bio.usage);
return result;
}
#elif LINUX_VERSION_CODE >=KERNEL_VERSION(2,6,0)
typedef void (* fn_submit_bio)(int,struct bio *);
blk_qc_t submit_bio_knook(int rw,struct bio * bio)
{
int rcod=0;
struct lbb_device * xdev;
fn_submit_bio fn_orig=(fn_submit_bio)submit_bio_orig;
atomic_inc(&k_submit_bio.usage);
if (rw & WRITE)
{
xdev=lbb_find_table(bio->bi_bdev);
if (xdev !=NULL)
{
rcod=lbb_check_block_bio_changed(xdev,bio);
}
}
if (rcod==0) fn_orig(rw,bio);
atomic_dec(&k_submit_bio.usage);
return 0;
}
blk_qc_t khook_submit_bio_origin(struct bio * bio,unsigned int opf)
{
int rw;
fn_submit_bio fn_orig=(fn_submit_bio)submit_bio_orig;
atomic_inc(&k_submit_bio.usage);
if (opf & REQ_OP_WRITE) rw=WRITE;
else rw=READ;
fn_orig(rw,bio);
atomic_dec(&k_submit_bio.usage);
return 0;
}
#endif
knook k_submit_bio=
{
.offset=HOOK_SET,.target=HOOK_HEX,.fn_hook=submit_bio_knook,.fn_orig=submit_bio_orig,};
