#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/semaphore.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/time.h>
#include <linux/uaccess.h>
#include "lbb_kernel.h"
#include "lbb_cdev_debug.h"
#include "errors.h"
#include "info.h"
#ifndef INFO_TIMER_TYPE
#define INFO_TIMER_TYPE (0)
#warning Kernel timers is disabled
#endif
#define TMP_FILE_MODE (00400)
#define TMP_FILE_ATTR (O_RDWR | O_EXCL | O_DIRECT)
unsigned long get_xsecond(void)
{
return get_seconds();
}
unsigned long x_copy_to_user(void * dst,const void * src,unsigned long size)
{
#if (INFO_COPY_TO_USER==1)
return copy_to_user(dst,src,size);
#elif (INFO_COPY_TO_USER==2)
return raw_copy_to_user(dst,src,size);
#else
#error Undefined copy_to_user function.
#endif
}
void x_bio_free_all(struct bio * bio)
{
if (bio==NULL) return;
#if (INFO_BIO_FOR_EACH==1)
{
struct bio_vec bvec;
int i;
bio_for_each_segment(&bvec,bio,i)
__free_page(bvec->bv_page);
xkfree(bio);
}
#elif (INFO_BIO_FOR_EACH==2)
{
struct bio_vec bvec;
int i;
bio_for_each_segment_all(&bvec,bio,i)
__free_page(bvec->bv_page);
xkfree(bio);
}
#elif (INFO_BIO_FOR_EACH==3)
{
bio_put(bio);
}
#else
#error Undefined macro bio_for_each.
#endif
}
#if (INFO_BIO_SPLIT==1)
struct bio * x_bio_split(struct bio * bio,int sectors,struct bio_set * bs)
{
struct bio * bio2=bio_split(bio,sectors,GFP_KERNEL,bs);
if (bio2 !=NULL) bio_chain(bio2,bio);
return bio2;
}
#elif (INFO_BIO_SPLIT==2)
struct bio * x_bio_split(struct bio * bio,int sect,struct bio_set * bs)
{
struct bio * bio2=bio_clone_bioset(bio,GFP_KERNEL,bs);
if (bio2 !=NULL)
{
bio_trim(bio2,0,sect);
bio_advance(bio,bio2->bi_size);
bio_chain(bio2,bio);
}
return bio2;
}
#else
#error Undefined bio_split function.
#endif
struct bio_set * x_bioset_create(void)
{
struct bio_set * bset=0L;
#if (INFO_BIOSET_CREATE==1)
bset=bioset_create(BIO_POOL_SIZE,0,BIOSET_NEED_BVECS);
#elif (INFO_BIOSET_CREATE==2)
bset=bioset_create(BIO_POOL_SIZE,0);
#elif (INFO_BIOSET_CREATE==3)
bset=kzalloc(sizeof(struct bio_set),GFP_KERNEL);
if (bset)
{
if (bioset_init(bset,BIO_POOL_SIZE,0,BIOSET_NEED_BVECS))
{
kfree(bset);
bset=0L;
}
}
#else
#error Undefined bioset_create type.
#endif
return bset;
}
void x_bioset_free(struct bio_set * bset)
{
if (bset==0L) return;
#if (INFO_BIOSET_CREATE==1)
bioset_free(bset);
#elif (INFO_BIOSET_CREATE==2)
bioset_free(bset);
#elif (INFO_BIOSET_CREATE==3)
bioset_exit(bset);
kfree(bset);
#else
#error Undefined bioset_create type.
#endif
}
void x_bio_reset(struct bio * bio)
{
#if (INFO_BIO_RESET==1)
bio_reset(bio);
#elif (INFO_BIO_RESET==2)
bio_init(bio);
#else
#error Undefined bio_reset function.
#endif
}
void x_bio_set_sector(struct bio * bio,sector_t sector)
{
#if (INFO_BIO_SECT_TYPE==1)
bio->bi_sector=sector;
#elif (INFO_BIO_SECT_TYPE==2)
bio->bi_iter.bi_sector=sector;
#else
#error Undefined struct bio sector type.
#endif
}
sector_t x_bio_get_sector(const struct bio * bio)
{
#if (INFO_BIO_SECT_TYPE==1)
return bio->bi_sector;
#elif (INFO_BIO_SECT_TYPE==2)
return bio->bi_iter.bi_sector;
#else
#error Undefined struct bio sector type.
#endif
}
void x_bio_copy_dev(struct bio * dst,struct bio * src)
{
#if (INFO_BIO_TYPE==2)
dst->bi_disk=src->bi_disk;
dst->bi_partno=src->bi_partno;
#elif (INFO_BIO_TYPE==1)
dst->bi_bdev=src->bi_bdev;
#else
#error Undefined struct bio type.
#endif
}
void x_bio_set_dev(struct bio * dst,struct block_device * bdev)
{
#if (INFO_BIO_TYPE==2)
dst->bi_disk=bdev->bd_disk;
dst->bi_partno=bdev->bd_partno;
#elif (INFO_BIO_TYPE==1)
dst->bi_bdev=bdev;
#else
#error Undefined struct bio type.
#endif
}
struct gendisk * x_bio_get_disk(struct bio * bio)
{
#if (INFO_BIO_TYPE==2)
return bio->bi_disk;
#elif (INFO_BIO_TYPE==1)
return bio->bi_bdev->bd_disk;
#else
#error Undefined struct bio type.
#endif
}
void x_bio_abort(struct bio * bio,int errnum)
{
#if (INFO_BIO_ENDIO==1)
bio_io_error(bio);
#elif (INFO_BIO_ENDIO==2)
bio_endio(bio,errnum);
#elif (INFO_BIO_ENDIO==3)
bio->bi_error=errnum;
bio_endio(bio);
#else
#error Undefined bio endio type.
#endif
}
#if (INFO_VFS_TYPE==2)
static int vfs_fallocate(struct file *file,int mode,loff_t offset,loff_t len)
{
struct inode *inode=file->f_path.dentry->d_inode;
if (offset < 0 || len <=0)
return-EINVAL;
if (!(file->f_mode & FMODE_WRITE))
return-EBADF;
if (S_ISFIFO(inode->i_mode))
return-ESPIPE;
if (!S_ISREG(inode->i_mode) && !S_ISDIR(inode->i_mode))
return-ENODEV;
if (((offset+len) > inode->i_sb->s_maxbytes) || ((offset+len) < 0))
return-EFBIG;
if (!inode->i_op->fallocate)
return-EOPNOTSUPP;
return inode->i_op->fallocate(inode,mode,offset,len);
}
#endif
int lbb_store_create_file(const char * fname,size_t file_size,struct inode ** inode,struct file ** filp)
{
struct file * filp_tmp=NULL;
int rcod=NO_ERR;
if ((inode==NULL)||(filp==NULL)) return ERR_ZERO_POINTER;
filp_tmp=filp_open(fname,O_CREAT | TMP_FILE_ATTR,TMP_FILE_MODE);
if (IS_ERR_OR_NULL(filp_tmp))
{
rcod=(int)PTR_ERR(filp_tmp);
if (rcod==-EEXIST)
{
PrintLog(0,KERN_INFO "File <%s> exist. Will be rewrited.\n",fname);
filp_tmp=filp_open(fname,O_TRUNC | TMP_FILE_ATTR,TMP_FILE_MODE);
}
if (IS_ERR_OR_NULL(filp_tmp))
{
rcod=(int)PTR_ERR(filp_tmp);
PrintLog(0,KERN_INFO "Error open store file <%s>. %s\n",fname,error_to_text(rcod));
goto err_exit;
}
}
rcod=vfs_fallocate(filp_tmp,0,0,file_size);
if (rcod !=NO_ERR)
{
PrintLog(0,KERN_INFO "Error allocate store file <%s>. %s\n",fname,error_to_text(rcod));
goto err_exit;
}
*filp=filp_tmp;
#if (INFO_VFS_TYPE==1)
*inode=file_inode(*filp);
#elif (INFO_VFS_TYPE==2)
*inode=(*filp)->f_path.dentry->d_inode;
#else
#error Undefined vfs version.
#endif
return NO_ERR;
err_exit:
*filp=NULL;
*inode=NULL;
return rcod;
}
int lbb_store_remove_file(struct file * filp)
{
int rcod;
if (filp==NULL) return NO_ERR;
rcod=filp_close(filp,NULL);
dget(filp->f_path.dentry);
#if (INFO_VFS_TYPE==1)
rcod=vfs_unlink(d_inode(filp->f_path.dentry->d_parent),filp->f_path.dentry,NULL);
#elif (INFO_VFS_TYPE==2)
rcod=vfs_unlink(filp->f_path.dentry->d_parent->d_inode,filp->f_path.dentry);
#else
#error Undefined vfs version.
#endif
dput(filp->f_path.dentry);
return NO_ERR;
}
size_t lbb_store_bmap(struct inode * inode,size_t pos)
{
return bmap((struct inode *)inode,pos);
}
struct block_device * lbb_store_get_file_bdev(struct inode * inode)
{
if (inode==NULL) return NULL;
return inode->i_sb->s_bdev;
}
unsigned long xbdev_orig_mode(void) {return (FMODE_READ);}
unsigned long xbdev_save_mode(void) {return (FMODE_READ | FMODE_WRITE | FMODE_EXCL);}
void xblkdev_put(struct block_device * bdev,unsigned int mode) {blkdev_put(bdev,mode);}
int x_register_blkdev(unsigned int major,const char * name) {return register_blkdev(major,name);}
void x_unregister_blkdev(unsigned int major,const char * name) {unregister_blkdev(major,name);}
spin_lock_ptr x_spin_lock_create(const char * name)
{
spinlock_t * lock=kmalloc(sizeof(spinlock_t),GFP_KERNEL);
if (lock !=NULL) spin_lock_init(lock);
return (spin_lock_ptr)lock;
}
void x_spin_lock_destroy(spin_lock_ptr lock) {kfree(lock);}
void x_spin_lock(spin_lock_ptr lock,unsigned int info) {spin_lock((spinlock_t *)lock);}
void x_spin_unlock(spin_lock_ptr lock) {spin_unlock((spinlock_t *)lock);}
void x_spin_lock_irq(spin_lock_ptr lock,unsigned int info) {spin_lock_irq((spinlock_t *)lock);}
void x_spin_unlock_irq(spin_lock_ptr lock) {spin_unlock_irq((spinlock_t *)lock);}
int x_spin_trylock(spin_lock_ptr lock) {return spin_trylock((spinlock_t *)lock);}
sprw_lock_ptr x_rw_lock_create(const char * name)
{
rwlock_t * rwlk=kmalloc(sizeof(rwlock_t),GFP_KERNEL);
if (rwlk !=NULL) rwlock_init(rwlk);
return (sprw_lock_ptr)rwlk;
}
void x_rw_lock_destroy(sprw_lock_ptr rwlk) {kfree(rwlk);}
void x_rw_read_lock(sprw_lock_ptr rwlk) {read_lock((rwlock_t *)rwlk);}
void x_rw_read_unlock(sprw_lock_ptr rwlk) {read_unlock((rwlock_t *)rwlk);}
void x_rw_write_lock(sprw_lock_ptr rwlk) {write_lock((rwlock_t *)rwlk);}
void x_rw_write_unlock(sprw_lock_ptr rwlk) {write_unlock((rwlock_t *)rwlk);}
void * xkmalloc(size_t size) {return kmalloc(size,GFP_KERNEL);}
size_t xstrlen(const char * str) {return strlen(str);}
char * xstrcpy (char * dst,const char *src) {return strcpy(dst,src);}
char * xstrncpy (char *dst,const char *src,size_t len) {return strncpy(dst,src,len);}
void * xmemset(void * dst,int value,size_t len) {return memset(dst,value,len);}
void * xmemcpy(void *dst,const void *src,size_t len ) {return memcpy(dst,src,len);}
void xkfree(const void * ptr) {return kfree(ptr);}
int xstrncmp (const char * str1,const char * str2,size_t len)
{return strncmp(str1,str2,len);}
void * xkmem_cache_create(const char * name,unsigned int size)
{return kmem_cache_create(name,size,0,SLAB_HWCACHE_ALIGN,NULL);}
void xkmem_cache_destroy(void * cache) {kmem_cache_destroy((struct kmem_cache *)cache);}
void * xkmem_cache_zalloc(void * cache) {return kmem_cache_zalloc((struct kmem_cache *)cache,GFP_KERNEL);}
void xkmem_cache_free(void * cache,void * p) {kmem_cache_free((struct kmem_cache *)cache,p);}
void xatomic_inc(atomic_t * v) {atomic_inc(v);}
void xatomic_dec(atomic_t * v) {atomic_dec(v);}
int xatomic_read(atomic_t * v) {return atomic_read(v);}
void xatomic_set(atomic_t * v,int i) {atomic_set(v,i);}
void xatomic_add(int i,atomic_t * v) {atomic_add(i,v);}
void xatomic_sub(int i,atomic_t * v) {atomic_sub(i,v);}
int xatomic_dec_and_test(atomic_t *v){return atomic_dec_and_test(v);}
struct semaphore * x_sema_create(int value)
{
struct semaphore * sem;
sem=kmalloc(sizeof(struct semaphore),GFP_KERNEL);
if (sem) sema_init(sem,value);
return sem;
}
void x_sema_destroy(struct semaphore * sem)
{
if (sem) kfree(sem);
}
void x_down(struct semaphore * sem) {down(sem);}
void x_up(struct semaphore * sem) {up(sem);}
unsigned long xmsleep_interruptible(unsigned int msecs)
{
return msleep_interruptible(msecs);
}
x_timer * x_create_timer(void (* callback)(unsigned long),unsigned long data)
{
struct timer_list * timer;
timer=kmalloc(sizeof(struct timer_list),GFP_KERNEL);
#if (INFO_TIMER_TYPE==1)
setup_timer(timer,callback,data);
#elif (INFO_TIMER_TYPE==2)
timer_setup(timer,(void (*)(struct timer_list *))callback,data);
#else
timer=0L;
#endif
return timer;
}
void x_destroy_timer(x_timer * timer)
{
if (timer==0L) return;
#if (INFO_TIMER_TYPE==1)
del_timer_sync(timer);
#elif (INFO_TIMER_TYPE==2)
del_timer_sync(timer);
#endif
kfree(timer);
}
void x_set_timer(x_timer * timer,unsigned long expires)
{
if (timer==0L) return;
#if (INFO_TIMER_TYPE==1)
mod_timer(timer,jiffies+expires);
#elif (INFO_TIMER_TYPE==2)
mod_timer(timer,jiffies+expires);
#endif
}
int xsprintf( char * str,const char * format,... )
{
int len;
va_list va;
va_start(va,format);
len=vsprintf(str,format,va);
va_end(va);
return len;
}
void PrintLog(int level,const char * format,...)
{
ssize_t len;
va_list va;
char buf[512];
if ((format==NULL)||(format[0]==0)) return;
if (level) {}
va_start(va,format);
len=vsnprintf(buf,500,format,va);
va_end(va);
printk(buf);
lbb_cdev_set(buf,len);
}
struct block_device * lbb_blockdev_open(char * dev_path,unsigned long _mode,void * holder)
{
int part_no,rcod;
struct block_device * bdev;
uint64 nr_sects;
fmode_t mode=(fmode_t)_mode;
#if (INFO_LOOKUP_BDEV==1)
bdev=lookup_bdev(dev_path);
#elif (INFO_LOOKUP_BDEV==2)
bdev=lookup_bdev(dev_path,0);
#else
#error Undefined lookup_bdev type.
#endif
if (IS_ERR_OR_NULL(bdev))
{
PrintLog(0,KERN_ERR "Error opening block device <%s> <%lu>\n",dev_path,PTR_ERR(bdev));
return NULL;
}
if (!bdget(bdev->bd_dev))
{
PrintLog(0,KERN_ERR "Error bdget() <%s>\n",dev_path);
return NULL;
}
#if (INFO_GET_BDEV_TYPE==1)
rcod=blkdev_get(bdev,mode,holder);
#elif (INFO_GET_BDEV_TYPE==2)
(void)holder;
rcod=blkdev_get(bdev,mode);
#else
#error Undefined blkdev_get function.
#endif
if (rcod)
{
PrintLog(0,KERN_ERR "Error blkdev_get() <%s>\n",dev_path);
bdput(bdev);
return NULL;
}
nr_sects=(uint64)bdev->bd_part->nr_sects;
#if (INFO_BIO_TYPE==2)
part_no=bdev->bd_partno;
#else
part_no=bdev->bd_part->partno;
#endif
PrintLog(0,KERN_INFO "Device[%u:%u] <%s> open. Size %llu sectors (%s),block size %u,partno %d\n",MAJOR(bdev->bd_dev),MINOR(bdev->bd_dev),dev_path,nr_sects,format_mem_size(nr_sects << SECTOR_SHIFT),bdev->bd_block_size,part_no);
return bdev;
}
int check_is_directory(const char * path)
{
struct file * filp=filp_open(path,O_RDONLY | O_DIRECTORY,(00400));
if (IS_ERR_OR_NULL(filp)) return-1;
filp_close(filp,NULL);
return 0;
}
int check_is_blockdev(const char * dev_path)
{
struct block_device * bdev;
#if (INFO_LOOKUP_BDEV==1)
bdev=lookup_bdev(dev_path);
#elif (INFO_LOOKUP_BDEV==2)
bdev=lookup_bdev(dev_path,0);
#else
#error Undefined lookup_bdev type.
#endif
if (IS_ERR_OR_NULL(bdev)) return-1;
return 0;
}
unsigned char avl_execute_func(unsigned char (* func)(void *),void * n)
{
if (func==NULL) return 0;
return func(n);
}
struct block_device * bdev_get_bd_contains(struct block_device * bdev)
{
if (bdev==NULL) return NULL;
return bdev->bd_contains;
}
size_t bdev_get_start_sect(struct block_device * bdev)
{
if (bdev==NULL) return 0;
return bdev->bd_part->start_sect;
}
size_t bdev_get_block_size(struct block_device * bdev)
{
if (bdev==NULL) return 0;
return bdev->bd_block_size;
}
sector_t bdev_get_nr_sects(struct block_device * bdev)
{
if (bdev==NULL) return 0;
if (bdev->bd_part==NULL) return 0;
return bdev->bd_part->nr_sects;
}
