#include <linux/bio.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include "lbb_kernel.h"
#include "lbb_mapper.h"
#include "lbb_device2.h"
#include "lbb_device.h"
#include "kernel_hook.h"
#include "errors.h"
#include "info.h"
#define MAP_DEBUG
#define TIMEOUT_BIO_REQ (10000)
#define TIMER_PERIOD (100)
#define ITST_PROC_INIT (0)
#define ITST_WAIT_READ (1)
#define ITST_QUEUE_READ (2)
#define ITST_WAIT_WRITE (3)
#define ITST_QUEUE_WRITE (4)
#define ITST_WAIT_FREE (5)
#define ISTX_NO_USED (0)
#define ISTX_WORK (1)
#define ISTX_SUCCESS (2)
#define ISTX_WAIT_FREE (3)
#define ISD_FREE (0)
#define ISD_DELETE (1)
#define ISD_TIMEOUT (2)
struct lbb_save_block
{
struct bio * wr_bio;
unsigned int stx;
sector_t sect_orig;
sector_t sect_save;
};
struct lbb_save_item
{
struct list_head check_item;
struct lbb_save_item *item_next;
unsigned int queue_flg;
struct lbb_device * xdev;
struct bio * rd_bio;
struct bio * df_bio;
atomic_t xcnt;
unsigned int stat;
unsigned int error;
unsigned int is_del;
unsigned long jiff;
unsigned int vcnt;
};
struct item_list
{
struct lbb_save_item * head;
struct lbb_save_item * tail;
};
atomic_t check_count;
struct list_head check_list;
spin_lock_ptr check_lock=NULL;
struct task_struct * it_thread=NULL;
struct bio_set * it_bset=NULL;
struct item_list it_list;
wait_queue_head_t it_head;
atomic_t it_cache_count;
unsigned long check_it_time;
x_timer * check_timer=NULL;
#ifdef MAP_DEBUG
struct mapper_debug
{
unsigned long is_wait;
unsigned long is_wait2;
unsigned long tmt_cnt;
unsigned long cnt_inp;
unsigned long cnt_out;
} mdbg;
#endif
unsigned int lbb_mapper_sys_info(char * buf)
{
char * fix=buf;
#ifdef MAP_DEBUG
unsigned long cnt_out=mdbg.cnt_out;
unsigned long cnt_inp=mdbg.cnt_inp;
#endif
sprintf(buf,"Map cache: %u ",atomic_read(&it_cache_count));
buf+=19;
buf+=sprintf(buf,"Req: %u \n",atomic_read(&check_count));
#ifdef MAP_DEBUG
sprintf(buf,"Timeout cnt:%lu ",mdbg.tmt_cnt);
buf+=19;
buf+=sprintf(buf,"Wait:%lu \n",mdbg.is_wait);
sprintf(buf,"Debug input:%lu ",cnt_inp);
buf+=19;
sprintf(buf,"out:%lu ",cnt_out);
buf+=11;
buf+=sprintf(buf,"queue:%ld \n",cnt_inp-cnt_out);
#endif
return buf-fix;
}
static void lbb_mapper_check_add_item(struct lbb_save_item * it)
{
it->jiff=jiffies;
list_add(&it->check_item,&check_list);
atomic_inc(&check_count);
}
static void lbb_mapper_check_free_item(struct lbb_save_item * it)
{
list_del(&it->check_item);
atomic_dec(&check_count);
}
static inline int lbb_item_queue_empty(void)
{
return (it_list.head==NULL);
}
static inline struct lbb_save_item * lbb_get_item_from_queue(void)
{
struct lbb_save_item * it;
it=it_list.head;
if (it)
{
it_list.head=it->item_next;
if (!it_list.head) it_list.tail=NULL;
it->item_next=NULL;
it->queue_flg=0;
}
return it;
}
static void lbb_add_item_to_queue(struct lbb_save_item * it)
{
if (it->queue_flg==0)
{
it->item_next=NULL;
if (it_list.tail) it_list.tail->item_next=it;
else it_list.head=it;
it_list.tail=it;
it->queue_flg=1;
}
wake_up(&it_head);
}
static void lbb_timer_check_it(struct lbb_save_item * it,unsigned long jiff)
{
if (it->is_del !=ISD_FREE)
{
if (atomic_read(&it->xcnt)==0)
{
lbb_add_item_to_queue(it);
}
else if ((jiff-it->jiff) > check_it_time)
{
if (it->is_del !=ISD_TIMEOUT)
{
it->is_del=ISD_TIMEOUT;
lbb_add_item_to_queue(it);
#ifdef MAP_DEBUG
mdbg.tmt_cnt++;
#endif
}
}
return;
}
if (it->error !=0) return;
if ((jiff-it->jiff) > check_it_time)
{
it->error=ERR_IT_WORK_TIMEOUT;
lbb_add_item_to_queue(it);
#ifdef MAP_DEBUG
mdbg.tmt_cnt++;
#endif
return;
}
if (it->stat==ITST_WAIT_FREE)
{
#ifdef MAP_DEBUG
mdbg.is_wait2++;
#endif
it->jiff=jiff;
it->stat=ITST_QUEUE_WRITE;
lbb_add_item_to_queue(it);
return;
}
}
static void lbb_timer_callback(unsigned long __val)
{
struct list_head * pos;
struct lbb_save_item * it;
unsigned long jiff=jiffies;
#ifdef MAP_DEBUG
mdbg.is_wait2=0;
#endif
x_spin_lock_irq(check_lock,231);
list_for_each(pos,&check_list)
{
it=(struct lbb_save_item *)pos;
lbb_timer_check_it(it,jiff);
}
x_spin_unlock_irq(check_lock);
#ifdef MAP_DEBUG
mdbg.is_wait=mdbg.is_wait2;
#endif
x_set_timer(check_timer,jiff+msecs_to_jiffies(TIMER_PERIOD));
}
static struct lbb_save_item * lbb_create_it(unsigned int vcnt)
{
unsigned int size,i;
struct lbb_save_block * sbl;
struct lbb_save_item * it;
size=sizeof(struct lbb_save_item)+sizeof(struct lbb_save_block) * vcnt;
it=xkmalloc(size);
if (it==NULL) return NULL;
memset(it,0,size);
it->is_del=ISD_FREE;
atomic_inc(&it_cache_count);
it->vcnt=vcnt;
atomic_set(&it->xcnt,0);
sbl=(struct lbb_save_block *)(it+1);
for (i=0;i < vcnt;++i)
{
sbl[i].sect_orig=ERR_SECT_NUM;
sbl[i].sect_save=ERR_SECT_NUM;
}
return it;
}
static void lbb_kill_it(struct lbb_save_item * it)
{
unsigned int i;
struct lbb_save_item * it_tmp=NULL;
struct lbb_save_block * sbl;
struct bio * rd_bio_tmp=NULL;
struct bio * df_bio_tmp=NULL;
if (it==NULL) return;
x_spin_lock_irq(check_lock,943);
if (it->is_del !=ISD_FREE)
{
it->is_del=ISD_DELETE;
it->jiff=jiffies;
}
rd_bio_tmp=it->rd_bio;
df_bio_tmp=it->df_bio;
it->rd_bio=NULL;
it->df_bio=NULL;
if ((atomic_read(&it->xcnt)==0)||(it->is_del==ISD_TIMEOUT))
{
lbb_mapper_check_free_item(it);
it_tmp=it;
}
x_spin_unlock_irq(check_lock);
if (it_tmp)
{
sbl=(struct lbb_save_block *)(it_tmp+1);
for (i=0;i < it_tmp->vcnt;++i)
x_bio_free_all(sbl[i].wr_bio);
atomic_dec(&it_cache_count);
xkfree(it_tmp);
}
if (rd_bio_tmp !=NULL)
{
xkfree(rd_bio_tmp);
}
if (df_bio_tmp !=NULL) khook_submit_bio_origin(df_bio_tmp,REQ_OP_WRITE);
}
int lbb_check_block_bio_changed(struct lbb_device * xdev,struct bio * bio)
{
int i,rcod;
struct bio_vec * bv;
struct lbb_save_item * it=NULL;
struct lbb_save_block * sbl=NULL;
sector_t sector,block,nsect;
if (!lbb_check_disk(xdev) || (lbb_dev_get_status(xdev) !=LBB_STAT_CATCH)) return 0;
nsect=0;
bv=bio->bi_io_vec;
sector=x_bio_get_sector(bio);
for (i=0;i < bio->bi_vcnt;++i)
{
sector+=nsect;
nsect=bv[i].bv_len >> SECTOR_SHIFT;
block=sector >> xdev->xbs;
if ((sector < xdev->sect_beg)||(sector >=xdev->sect_end))
continue;
if (lbb_map_check_status(xdev->skip,block)) continue;
if (!lbb_block_store_check(xdev->store,block)) continue;
if (it==NULL)
{
it=lbb_create_it(bio->bi_vcnt);
if (it==NULL)
{
rcod=ERR_NO_MEMORY;
PrintLog(0,KERN_ERR "LBB error lbb_save_item\n");
goto err_exit;
}
it->df_bio=bio;
it->xdev=xdev;
sbl=(struct lbb_save_block *)(it+1);
}
sbl[i].sect_orig=sector;
sbl[i].stx=ISTX_WORK;
}
if (it !=NULL)
{
#ifdef MAP_DEBUG
mdbg.cnt_inp++;
#endif
x_spin_lock_irq(check_lock,89);
lbb_mapper_check_add_item(it);
lbb_add_item_to_queue(it);
x_spin_unlock_irq(check_lock);
return 1;
}
err_exit:
return 0;
}
#if (INFO_BIO_END_IO==1)
static void lbb_bio_end_read_request(struct bio * bio)
#elif (INFO_BIO_END_IO==2)
static void lbb_bio_end_read_request(struct bio * bio,int parm)
#else
#error Undefined bio_end_io_t type.
#endif
{
struct lbb_save_item * it=bio->bi_private;
if (it==NULL)
{
PrintLog(0,KERN_ERR "Error bio at end_read_request\n");
return;
}
x_spin_lock_irq(check_lock,36);
if (it->stat !=ITST_WAIT_READ) it->error=ERR_IT_READ_REQUEST;
else it->stat=ITST_QUEUE_READ;
it->jiff=jiffies;
atomic_dec(&it->xcnt);
lbb_add_item_to_queue(it);
x_spin_unlock_irq(check_lock);
}
#if (INFO_BIO_END_IO==1)
static void lbb_bio_end_write_request(struct bio * bio)
#elif (INFO_BIO_END_IO==2)
static void lbb_bio_end_write_request(struct bio * bio,int parm)
#else
#error Undefined bio_end_io_t type.
#endif
{
int i,cnt;
struct lbb_save_block * sbl;
struct lbb_save_item * it=bio->bi_private;
if (it==NULL)
{
PrintLog(0,KERN_ERR "Error bio at end_write_request\n");
return;
}
x_spin_lock_irq(check_lock,99);
if (it->stat !=ITST_WAIT_WRITE)
{
it->error=ERR_IT_WRITE_REQUEST;
lbb_add_item_to_queue(it);
goto exit;
}
sbl=(struct lbb_save_block *)(it+1);
for (cnt=0,i=0;i < it->vcnt;++i)
{
if (sbl[i].wr_bio==bio) sbl[i].stx=ISTX_SUCCESS;
else if (sbl[i].stx==ISTX_WORK) cnt++;
}
if (cnt==0)
{
it->stat=ITST_QUEUE_WRITE;
lbb_add_item_to_queue(it);
goto exit;
}
exit:
atomic_dec(&it->xcnt);
x_spin_unlock_irq(check_lock);
}
static int lbb_mapper_work_thread(void * data)
{
int rcod,i,cnt;
sector_t bsave,borig,block;
struct lbb_save_item * it;
struct lbb_save_block * sbl;
struct bio_vec * bv;
struct page * page;
PrintLog(0,KERN_NOTICE "Run Thread queue save to store\n");
set_user_nice(current,5);
while (1)
{
wait_event_interruptible(it_head,kthread_should_stop() || !lbb_item_queue_empty());
if (kthread_should_stop()) break;
x_spin_lock_irq(check_lock,26);
it=lbb_get_item_from_queue();
x_spin_unlock_irq(check_lock);
if (it==NULL) continue;
if ((it->is_del !=ISD_FREE) ||
!lbb_check_disk(it->xdev) ||
(lbb_dev_get_status(it->xdev) !=LBB_STAT_CATCH))
{
lbb_kill_it(it);
continue;
}
if (it->df_bio==NULL) {rcod=ERR_ZERO_POINTER;goto err_exit;}
if (it->error !=NO_ERR) {rcod=it->error;goto err_exit;}
if (it->stat==ITST_PROC_INIT)
{
it->rd_bio=bio_alloc_bioset(GFP_KERNEL,it->df_bio->bi_vcnt,it_bset);
if (it->rd_bio==NULL)
{
rcod=ERR_CREATE_RD_BIO;
goto err_exit;
}
x_bio_reset(it->rd_bio);
x_bio_copy_dev(it->rd_bio,it->df_bio);
x_bio_set_sector(it->rd_bio,x_bio_get_sector(it->df_bio));
sbl=(struct lbb_save_block *)(it+1);
bv=it->df_bio->bi_io_vec;
cnt=0;
for (i=0;i < it->vcnt;++i)
{
page=alloc_page(GFP_KERNEL);
if (!page)
{
rcod=ERR_ALLOC_PAGE;
goto err_exit;
}
bio_add_page(it->rd_bio,page,bv[i].bv_len,bv[i].bv_offset);
if (sbl[i].stx==ISTX_WORK)
{
cnt++;
sbl[i].sect_save=lbb_block_store_allocate(it->xdev->store);
if (sbl[i].sect_save==ERR_SECT_NUM)
{
rcod=ERR_STORE_IS_FULL;
goto err_exit;
}
if (it->xdev->bdev_orig==lbb_block_store_get_dev(it->xdev->store))
{
block=sbl[i].sect_save >> it->xdev->xbs;
rcod=lbb_map_set_range(it->xdev->skip,block,1,1,BM_SET);
}
}
}
if (cnt==0)
{
rcod=ERR_IT_ZERO_REQUEST;
goto err_exit;
}
it->rd_bio->bi_private=it;
it->rd_bio->bi_end_io=lbb_bio_end_read_request;
it->stat=ITST_WAIT_READ;
atomic_inc(&it->xcnt);
khook_submit_bio_origin(it->rd_bio,REQ_OP_READ);
continue;
}
if (it->stat==ITST_QUEUE_READ)
{
if (it->rd_bio==NULL)
{
rcod=ERR_INTERNAL4;
goto err_exit;
}
sbl=(struct lbb_save_block *)(it+1);
bv=it->rd_bio->bi_io_vec;
for (i=0;i < it->vcnt;++i)
if (sbl[i].stx==ISTX_WORK)
{
sbl[i].wr_bio=bio_alloc_bioset(GFP_KERNEL,2,it_bset);
if (sbl[i].wr_bio==NULL)
{
rcod=ERR_CREATE_WR_BIO;
goto err_exit;
}
x_bio_reset(sbl[i].wr_bio);
x_bio_set_dev(sbl[i].wr_bio,lbb_block_store_get_dev(it->xdev->store));
x_bio_set_sector(sbl[i].wr_bio,sbl[i].sect_save);
sbl[i].wr_bio->bi_private=it;
sbl[i].wr_bio->bi_end_io=lbb_bio_end_write_request;
if (bv[i].bv_page==NULL)
{
rcod=ERR_INTERNAL3;
goto err_exit;
}
bio_add_page(sbl[i].wr_bio,bv[i].bv_page,bv[i].bv_len,bv[i].bv_offset);
bv[i].bv_page=NULL;
bv[i].bv_len=0;
}
xkfree(it->rd_bio);
it->rd_bio=NULL;
it->jiff=jiffies;
it->stat=ITST_WAIT_WRITE;
for (i=0;i < it->vcnt;++i)
if (sbl[i].stx==ISTX_WORK)
{
if (sbl[i].wr_bio==NULL)
{
rcod=ERR_INTERNAL2;
goto err_exit;
}
atomic_inc(&it->xcnt);
khook_submit_bio_origin(sbl[i].wr_bio,REQ_OP_WRITE);
}
continue;
}
if (it->stat==ITST_QUEUE_WRITE)
{
rcod=0;
cnt=0;
sbl=(struct lbb_save_block *)(it+1);
for (i=0;i < it->vcnt;++i)
{
if ((sbl[i].stx==ISTX_SUCCESS)||(sbl[i].stx==ISTX_WAIT_FREE))
{
if (!lbb_remap_check_range(sbl[i].sect_orig))
{
sbl[i].stx=ISTX_NO_USED;
bsave=sbl[i].sect_save >> it->xdev->xbs;
borig=sbl[i].sect_orig >> it->xdev->xbs;
rcod=lbb_block_store_activate(it->xdev->store,bsave,borig);
if (rcod !=0) break;
}
else
{
sbl[i].stx=ISTX_WAIT_FREE;
cnt++;
}
}
}
if (rcod !=0) goto err_exit;
if (cnt==0)
{
lbb_kill_it(it);
#ifdef MAP_DEBUG
mdbg.cnt_out++;
#endif
}
else
{
it->jiff=jiffies;
it->stat=ITST_WAIT_FREE;
}
continue;
}
PrintLog(0,KERN_ERR "Error unknown item status\n");
rcod=ERR_IT_UNKNOWN_STAT;
goto err_exit;
err_exit:
lbb_stop_catch(it->xdev,rcod);
PrintLog(0,KERN_ERR "Mapper queue error. %s\n",error_to_text(rcod));
if (it !=NULL) PrintLog(0,KERN_ERR "Item stat=%d vcnt=%d %s @%p\n",it->stat,it->vcnt,error_to_text(it->error),it);
lbb_kill_it(it);
continue;
}
PrintLog(0,KERN_NOTICE "Exit Thread queue save to store\n");
return 0;
}
int lbb_mapper_init(struct bio_set * bset)
{
if (bset==NULL) return ERR_ZERO_POINTER;
it_bset=bset;
atomic_set(&it_cache_count,0);
#ifdef MAP_DEBUG
memset(&mdbg,0,sizeof(struct mapper_debug));
#endif
check_lock=x_spin_lock_create("mapper check_lock");
atomic_set(&check_count,0);
INIT_LIST_HEAD(&check_list);
init_waitqueue_head(&it_head);
it_list.head=it_list.tail=NULL;
it_thread=kthread_create(lbb_mapper_work_thread,NULL,"cblb.mapper");
if (IS_ERR_OR_NULL(it_thread)) return-1;
check_it_time=msecs_to_jiffies(TIMEOUT_BIO_REQ);
check_timer=x_create_timer(lbb_timer_callback,0);
x_set_timer(check_timer,jiffies+msecs_to_jiffies(TIMER_PERIOD+TIMER_PERIOD));
wake_up_process(it_thread);
return 0;
}
int lbb_mapper_free_disk(struct lbb_device * xdev)
{
int cnt=0;
struct lbb_save_item * it;
if (xdev==NULL) return 0;
if (lbb_dev_get_status(xdev)==LBB_STAT_CATCH)
{
lbb_stop_catch(xdev,ERR_UNSTOP_DEV);
PrintLog(0,KERN_ERR "Mapper error. Free unstopped xdev\n");
}
while (1)
{
x_spin_lock_irq(check_lock,28);
it=it_list.head;
while (1)
{
if (it==NULL) break;
if (it->xdev==xdev) break;
it=it->item_next;
}
x_spin_unlock_irq(check_lock);
if (it==NULL) break;
if (++cnt > 20)
{
PrintLog(0,KERN_ERR "Mapper error. Timeout wait free queue\n");
return ERR_UNSTOP_DEV;
}
PrintLog(0,KERN_INFO "Mapper stop loop %d.\n",cnt);
msleep_interruptible(100);
}
return 0;
}
void lbb_mapper_destroy(void)
{
int tmp,wait_cnt=0;
while (atomic_read(&check_count) !=0)
{
if (++wait_cnt > 100)
{
PrintLog(0,KERN_ERR "Error wait map_count=%u\n",atomic_read(&check_count));
break;
}
msleep_interruptible(100);
}
x_destroy_timer(check_timer);
check_timer=NULL;
if (it_thread !=NULL)
{
wake_up_process(it_thread);
kthread_stop(it_thread);
}
{
tmp=atomic_read(&it_cache_count);
if (tmp !=0) PrintLog(0,KERN_ERR "Error. Unfree %d lbb_mapper items\n",tmp);
}
x_spin_lock_destroy(check_lock);
check_lock=NULL;
}
