#include <linux/version.h>
#include <linux/kthread.h>
#include <linux/blkdev.h>
#include <linux/bio.h>
#include <linux/hdreg.h>
#include <linux/delay.h>
#include "errors.h"
#include "lbb_kernel.h"
#include "lbb_device.h"
#include "lbb_device2.h"
#include "lbb_block_store.h"
#include "kernel_hook.h"
#include "lbb_remap.h"
#include "info.h"
#define RMST_PROC_BEGIN (0)
#define RMST_PROC_WAIT (1)
#define RMST_PROC_END (2)
struct _lbb_remap_dev
{
atomic_t open_count;
struct block_device * bdev_orig;
spin_lock_ptr store_lock;
lbb_block_store * store;
lbb_map_tree * skip;
unsigned char xbs;
struct request_queue * queue;
struct gendisk * disk;
};
struct lbb_remap_item
{
struct list_head check_item;
struct lbb_remap_item * item_next;
lbb_remap_dev * rdev;
struct bio * bio;
unsigned int stat;
unsigned int error;
unsigned int queue_flg;
void * bi_end_io;
void * bi_private;
sector_t sect_beg;
sector_t sect_end;
};
struct rm_item_list
{
struct lbb_remap_item * head;
struct lbb_remap_item * tail;
};
unsigned int wait_count;
struct list_head wait_list;
spin_lock_ptr wait_lock=NULL;
struct task_struct * rm_thread=NULL;
struct bio_set * rm_bset=NULL;
struct rm_item_list rm_list_it;
struct bio_list rm_list_bio;
spin_lock_ptr rm_bio_lock=NULL;
spin_lock_ptr rm_it_lock=NULL;
wait_queue_head_t rm_head;
struct kmem_cache * rm_cache=NULL;
spin_lock_ptr rm_cache_lock;
atomic_t rm_cache_count;
unsigned int rm_bio_count;
unsigned int lbb_remap_sys_info(char * buf)
{
unsigned int tmp,tmp2;
char * fix=buf;
tmp=atomic_read(&rm_cache_count);
x_spin_lock_irq(rm_bio_lock,415);
tmp2=rm_bio_count;
x_spin_unlock_irq(rm_bio_lock);
buf+=sprintf(buf,"Queue it: %u \t Queue bio: %u \n",tmp,tmp2);
return buf-fix;
}
static int fn_lbb_remap_getgeo(struct block_device * bdev,struct hd_geometry * geo)
{
lbb_remap_dev * rdev=bdev->bd_disk->private_data;
return rdev->bdev_orig->bd_disk->fops->getgeo(rdev->bdev_orig,geo);
}
static int fn_lbb_remap_ioctl(struct block_device * bdev,fmode_t mode,unsigned int cmd,unsigned long arg)
{
int rcod=-ENOTTY;
lbb_remap_dev * rdev=bdev->bd_disk->private_data;
if (rdev==NULL) return-ENOTTY;
if ((cmd & 0xff00)==0x5300) return-ENOTTY;
switch ( cmd )
{
case HDIO_GETGEO:
{
struct hd_geometry geo;
fn_lbb_remap_getgeo(bdev,&geo);
if (copy_to_user((void __user *)arg,&geo,sizeof(geo))) return-EFAULT;
rcod=0;
break;
}
default:
PrintLog(0,KERN_INFO "lbb_ioctl unknown command 0x%04x\n",cmd);
rcod=-ENOTTY;
break;
}
return rcod;
}
unsigned int lbb_remap_get_users(lbb_remap_dev * rdev)
{
if (rdev==NULL) return 0;
return atomic_read(&rdev->open_count);
}
static int fn_lbb_remap_open(struct block_device * bdev,fmode_t mode)
{
int rcod=0;
lbb_remap_dev * rdev=bdev->bd_disk->private_data;
if (mode & FMODE_WRITE) {rcod=-EROFS;goto out;}
if (rdev==NULL) {rcod=-ENXIO;goto out;}
atomic_inc(&rdev->open_count);
out:
return rcod;
}
static inline void __lbb_remap_release(struct gendisk * disk,fmode_t mode)
{
lbb_remap_dev * rdev;
(void)mode;
if (disk==NULL) return;
rdev=disk->private_data;
if (rdev==NULL) return;
atomic_dec(&rdev->open_count);
}
#if (INFO_DEV_OP==1)
static void fn_lbb_remap_release(struct gendisk * disk,fmode_t mode)
{
__lbb_remap_release(disk,mode);
return;
}
#elif (INFO_DEV_OP==2)
static int fn_lbb_remap_release(struct gendisk * disk,fmode_t mode)
{
__lbb_remap_release(disk,mode);
return 0;
}
#else
#error Undefined bio_split function.
#endif
static struct block_device_operations lbb_remap_ops=
{
.owner=THIS_MODULE,.ioctl=fn_lbb_remap_ioctl,.open=fn_lbb_remap_open,.release=fn_lbb_remap_release,.getgeo=fn_lbb_remap_getgeo,};
static inline void lbb_remap_add_item_to_queue(struct lbb_remap_item * it)
{
x_spin_lock_irq(rm_it_lock,26);
if (it->queue_flg==0)
{
it->item_next=NULL;
if (rm_list_it.tail) rm_list_it.tail->item_next=it;
else rm_list_it.head=it;
rm_list_it.tail=it;
it->queue_flg=1;
}
x_spin_unlock_irq(rm_it_lock);
wake_up(&rm_head);
}
static inline int lbb_remap_item_queue_empty(void)
{
return (rm_list_it.head==NULL);
}
static inline struct lbb_remap_item * lbb_remap_get_item_from_queue(void)
{
struct lbb_remap_item * it;
x_spin_lock_irq(rm_it_lock,75);
it=rm_list_it.head;
if (it)
{
rm_list_it.head=it->item_next;
if (!rm_list_it.head) rm_list_it.tail=NULL;
it->item_next=NULL;
it->queue_flg=0;
}
x_spin_unlock_irq(rm_it_lock);
return it;
}
static void lbb_remap_abort_bio(struct bio * bio,int errnum)
{
if (bio==NULL) return;
x_bio_abort(bio,errnum);
}
static blk_qc_t lbb_remap_queue_make_request(struct request_queue * q,struct bio * bio)
{
int rcod;
lbb_remap_dev * rdev=q->queuedata;
if (bio==NULL) {rcod=ERR_ZERO_POINTER;goto abort;}
if (rdev->store==NULL) {rcod=ERR_SNAPSHOT_OFF;goto abort;}
if (bio_data_dir(bio) !=0) {rcod=0;goto abort;}
x_spin_lock_irq(rm_bio_lock,45);
bio_list_add(&rm_list_bio,bio);
rm_bio_count++;
x_spin_unlock_irq(rm_bio_lock);
wake_up(&rm_head);
return BLK_QC_T_NONE;
abort:
lbb_remap_abort_bio(bio,rcod);
if (rcod !=0)
{
if (rcod==ERR_SNAPSHOT_OFF) rcod=-EBUSY;
else PrintLog(0,KERN_ERR "Remap queue request error %s\n",error_to_text(rcod));
}
return BLK_QC_T_NONE;
}
int lbb_remap_check_range(sector_t sect)
{
int rcod=0;
struct list_head * pos;
struct lbb_remap_item * it;
x_spin_lock(wait_lock,10);
list_for_each(pos,&wait_list)
{
it=(struct lbb_remap_item *)pos;
if ((sect >=it->sect_beg)&&(sect < it->sect_end))
{
rcod=1;
break;
}
}
x_spin_unlock(wait_lock);
return rcod;
}
const char * lbb_remap_disk_name(lbb_remap_dev * rdev)
{
static const char * str_err="NULL";
if ((rdev==NULL)||(rdev->disk==NULL)) return str_err;
return rdev->disk->disk_name;
}
struct gendisk * lbb_remap_disk_pntr(lbb_remap_dev * rdev)
{
if (rdev==NULL) return NULL;
return rdev->disk;
}
static void lbb_remap_free_disk_priv(lbb_remap_dev * rdev)
{
if (rdev==NULL) return;
if (rdev->disk !=NULL)
{
del_gendisk(rdev->disk);
put_disk(rdev->disk);
}
if (rdev->queue !=NULL)
{
blk_cleanup_queue(rdev->queue);
}
x_spin_lock_destroy(rdev->store_lock);
xkfree(rdev);
}
void lbb_remap_set_store(lbb_remap_dev * rdev,void * store)
{
if (rdev==NULL) return;
x_spin_lock(rdev->store_lock,11);
rdev->store=store;
x_spin_unlock(rdev->store_lock);
}
lbb_remap_dev * lbb_remap_open_disk(
struct block_device * bdev_orig,const char * snap_shot_name,sector_t capacity,lbb_map_tree * skip,unsigned char rl_to_sect,int major,int minor,int * rcod_p)
{
sector_t max_hw_sect;
lbb_remap_dev * rdev=NULL;
int rcod;
if ((bdev_orig==NULL)||(snap_shot_name==NULL))
{
rcod=ERR_PARAMS;
goto err_exit;
}
rdev=xkmalloc(sizeof(lbb_remap_dev));
if (rdev==NULL)
{
rcod=ERR_NO_MEMORY;
goto err_exit;
}
memset(rdev,0,sizeof(lbb_remap_dev));
rdev->bdev_orig=bdev_orig;
rdev->store_lock=x_spin_lock_create("rdev->store_lock");
rdev->skip=skip;
rdev->xbs=rl_to_sect;
rdev->store=NULL;
atomic_set(&rdev->open_count,0);
rdev->queue=blk_alloc_queue(GFP_KERNEL);
if (rdev->queue==NULL )
{
rcod=ERR_NO_MEMORY;
goto err_exit;
}
blk_queue_make_request(rdev->queue,lbb_remap_queue_make_request);
blk_queue_logical_block_size(rdev->queue,bdev_orig->bd_block_size);
rdev->queue->queuedata=rdev;
max_hw_sect=queue_max_hw_sectors(bdev_get_queue(bdev_orig));
blk_queue_max_hw_sectors(rdev->queue,max_hw_sect);
rdev->disk=alloc_disk(1);
if (! rdev->disk)
{
rcod=ERR_NO_MEMORY;
goto err_exit;
}
strncpy(rdev->disk->disk_name,snap_shot_name,16);
rdev->disk->disk_name[15]=0;
rdev->disk->major=major;
rdev->disk->minors=1;
rdev->disk->first_minor=minor;
rdev->disk->fops=&lbb_remap_ops;
rdev->disk->queue=rdev->queue;
rdev->disk->private_data=rdev;
set_capacity(rdev->disk,capacity);
add_disk(rdev->disk);
PrintLog(0,KERN_INFO "Create snap_shot device <%s> OK\n",snap_shot_name);
return rdev;
err_exit:
lbb_remap_free_disk_priv(rdev);
PrintLog(0,KERN_ERR "Error create snap_shot device <%s> rcod=%d\n",snap_shot_name,rcod);
if (rcod_p) *rcod_p=rcod;
return NULL;
}
void lbb_remap_free_disk(lbb_remap_dev * rdev)
{
lbb_remap_free_disk_priv(rdev);
PrintLog(0,KERN_INFO "Remove snap_shot device\n");
}
#if (INFO_BIO_END_IO==1)
static void lbb_remap_wait_cb(struct bio * bio)
{
struct lbb_remap_item * it=bio->bi_private;
if (it==NULL) return;
if (it->stat !=RMST_PROC_WAIT) it->error=ERR_IT_READ_REQUEST;
it->stat=RMST_PROC_END;
bio->bi_private=it->bi_private;
bio->bi_end_io=it->bi_end_io;
if (bio->bi_end_io !=NULL) bio->bi_end_io(bio);
lbb_remap_add_item_to_queue(it);
}
#elif (INFO_BIO_END_IO==2)
static void lbb_remap_wait_cb(struct bio * bio,int parm)
{
struct lbb_remap_item * it=bio->bi_private;
if (it==NULL) return;
if (it->stat !=RMST_PROC_WAIT) it->error=ERR_IT_READ_REQUEST;
it->stat=RMST_PROC_END;
bio->bi_private=it->bi_private;
bio->bi_end_io=it->bi_end_io;
if (bio->bi_end_io !=NULL) bio->bi_end_io(bio,parm);
lbb_remap_add_item_to_queue(it);
}
#else
#error Undefined bio_end_io_t type.
#endif
static int lbb_remap_add_wait_it(lbb_remap_dev * rdev,struct bio * bio,sector_t sect_beg,sector_t sect_end)
{
struct lbb_remap_item * it;
if ((rdev==NULL)||(bio==NULL))
return ERR_ZERO_POINTER;
it=kmem_cache_zalloc(rm_cache,GFP_KERNEL);
if (it !=NULL) atomic_inc(&rm_cache_count);
else return ERR_NO_MEMORY;
it->rdev=rdev;
it->queue_flg=0;
it->error=0;
it->sect_beg=sect_beg;
it->sect_end=sect_end;
it->bio=bio;
it->bi_private=bio->bi_private;
it->bi_end_io=bio->bi_end_io;
it->stat=RMST_PROC_WAIT;
bio->bi_private=it;
bio->bi_end_io=lbb_remap_wait_cb;
x_spin_lock(wait_lock,12);
list_add(&it->check_item,&wait_list);
wait_count++;
x_spin_unlock(wait_lock);
return NO_ERR;
}
static void lbb_remap_kill_it(struct lbb_remap_item * it,int err_code)
{
if (it==NULL) return;
if ((err_code !=0)&&(it->bio !=NULL))
bio_io_error(it->bio);
x_spin_lock(wait_lock,14);
list_del(&it->check_item);
wait_count--;
x_spin_unlock(wait_lock);
kmem_cache_free(rm_cache,it);
atomic_dec(&rm_cache_count);
}
static struct bio * lbb_remap_bio_split(struct bio * bio,sector_t sect)
{
struct bio * bio2;
sector_t beg,end;
beg=x_bio_get_sector(bio);
end=beg+bio_sectors(bio)-1;
if ((sect < beg)||(sect >=end)) goto err_exit;
bio2=x_bio_split(bio,(sect-x_bio_get_sector(bio)),rm_bset);
if (bio2==NULL) goto err_exit;
return bio2;
err_exit:
PrintLog(0,KERN_ERR "Error split bio: {beg:%06llx,sect:%06llx,end:%06llx}\n",beg,sect,end);
return NULL;
}
#ifdef ENABLE_FAKE_READ
static inline void lbb_remap_read_skip(struct bio * bio)
{
char * buffer;
bio_for_each_segment(bv,bio,iter)
{
buffer=__bio_kmap_atomic(bio,iter);
__bio_kunmap_atomic(buffer);
}
bio->bi_error=0;
bio_endio(bio);
PrintLog(0,KERN_NOTICE "[%u]SKIP: (%06lx..%06lx) [x]\n",deep,sect_beg,sector-1);
}
if (lbb_map_check_opt(map))
else if (lbb_map_check_skip(rdev->skip,sect_beg >> rdev->xbs))
{
is_orig=0;
bio2=lbb_remap_bio_split(bio,sector);
x_bio_set_sector(bio,sect_beg+sector);
lbb_remap_read_skip(bio);
bio=bio2;
goto loop;
}
#endif
static inline void lbb_remap_read_save(struct bio * bio,lbb_remap_dev * rdev,sector_t fsect)
{
x_bio_set_sector(bio,fsect);
x_bio_set_dev(bio,lbb_block_store_get_dev(rdev->store));
khook_submit_bio_origin(bio,REQ_OP_READ);
}
static inline void lbb_remap_read_orig(struct bio * bio,lbb_remap_dev * rdev)
{
int rcod;
sector_t sector=x_bio_get_sector(bio);
rcod=lbb_remap_add_wait_it(rdev,bio,sector,(sector+bio_sectors(bio)-1));
if (rcod) {}
x_bio_set_dev(bio,rdev->bdev_orig);
khook_submit_bio_origin(bio,REQ_OP_READ);
}
static inline int lbb_remap_make_request(lbb_remap_dev * rdev,struct bio * bio)
{
struct bio * bio2;
sector_t esect,sect,fsect=ERR_SECT_NUM;
int i,is_orig,is_rest;
struct bio_vec * bv;
loop:
if (bio==NULL) return ERR_BIO_POINTER;
sect=x_bio_get_sector(bio);
esect=sect+bio_sectors(bio)-1;
is_orig=is_rest=0;
for (i=0,bv=bio->bi_io_vec;i < bio->bi_vcnt;++i)
{
if (sect >=esect) break;
if (is_rest)
{
bio2=lbb_remap_bio_split(bio,sect);
if (bio2==NULL) return ERR_BIO_SPLIT;
lbb_remap_read_save(bio2,rdev,fsect);
goto loop;
}
x_spin_lock(rdev->store_lock,13);
if (rdev->store==NULL) fsect=ERR_SECT_NUM;
else fsect=lbb_block_store_map(rdev->store,sect);
x_spin_unlock(rdev->store_lock);
if (fsect==ERR_SECT_NUM)
{
is_orig++;
}
else
{
if (is_orig)
{
bio2=lbb_remap_bio_split(bio,sect);
if (bio2==NULL) return ERR_BIO_SPLIT;
lbb_remap_read_orig(bio2,rdev);
goto loop;
}
is_rest++;
}
sect+=(bv[i].bv_len >> SECTOR_SHIFT);
}
if (is_orig)
lbb_remap_read_orig(bio,rdev);
else if (is_rest && (fsect !=ERR_SECT_NUM))
lbb_remap_read_save(bio,rdev,fsect);
return NO_ERR;
}
static int lbb_remap_work_thread(void * data)
{
int rcod;
struct bio * bio;
lbb_remap_dev * rdev;
struct lbb_device * xdev;
struct lbb_remap_item * it;
PrintLog(0,KERN_NOTICE "Run Thread queue snap-shot read\n");
set_user_nice(current,-20);
while (1)
{
wait_event_interruptible(rm_head,kthread_should_stop() ||
!bio_list_empty(&rm_list_bio) ||
!lbb_remap_item_queue_empty());
if (kthread_should_stop()) break;
if (!bio_list_empty(&rm_list_bio))
{
rcod=0;
xdev=0;
rdev=0;
x_spin_lock_irq(rm_bio_lock,49);
bio=bio_list_pop(&rm_list_bio);
if (bio==NULL) rcod=-120;
else rm_bio_count--;
x_spin_unlock_irq(rm_bio_lock);
if (!rcod)
{
xdev=lbb_find_by_remap_dev(x_bio_get_disk(bio));
if (xdev==NULL) rcod=-121;
}
if (!rcod)
{
rdev=xdev->rmp_dev;
if (rdev==NULL) rcod=-122;
if (rdev->store==NULL) rcod=ERR_SNAPSHOT_OFF;
}
if (!rcod) rcod=lbb_remap_make_request(rdev,bio);
if (rcod)
{
PrintLog(0,KERN_ERR "Remap error next <bio> from queue rcod=%d\n",rcod);
lbb_remap_abort_bio(bio,rcod);
}
}
it=lbb_remap_get_item_from_queue();
if ((it==NULL)||(it->bio==NULL))
{
lbb_remap_kill_it(it,-2);
continue;
}
if (it->error !=0)
{
rcod=it->error;
PrintLog(0,KERN_ERR "Remap error item stat=%d error=%d @%p\n",it->stat,it->error,it);
lbb_remap_kill_it(it,rcod);
continue;
}
if (it->stat==RMST_PROC_END)
{
lbb_remap_kill_it(it,0);
continue;
}
PrintLog(0,KERN_ERR "Error unknown item status\n");
lbb_remap_kill_it(it,-4);
continue;
}
PrintLog(0,KERN_NOTICE "Exit Thread queue snap-shot read\n");
return 0;
}
int lbb_remap_init(struct bio_set * bset)
{
if (bset==NULL) return ERR_ZERO_POINTER;
rm_bset=bset;
atomic_set(&rm_cache_count,0);
rm_cache=kmem_cache_create("rm_cache",sizeof(struct lbb_remap_item),0,SLAB_HWCACHE_ALIGN,NULL);
wait_lock=x_spin_lock_create("remap wait_lock");
INIT_LIST_HEAD(&wait_list);
wait_count=0;
init_waitqueue_head(&rm_head);
rm_list_it.head=rm_list_it.tail=NULL;
rm_bio_lock=x_spin_lock_create("remap bio lock");
rm_bio_count=0;
bio_list_init(&rm_list_bio);
rm_it_lock=x_spin_lock_create("remap it lock");
rm_thread=kthread_create(lbb_remap_work_thread,NULL,"cblb.remap");
if (IS_ERR_OR_NULL(rm_thread)) return-1;
wake_up_process(rm_thread);
return 0;
}
void lbb_remap_destroy(void)
{
unsigned int cnt=0;
while (1)
{
x_spin_lock(wait_lock,17);
if (list_empty(&wait_list)) break;
x_spin_unlock(wait_lock);
if (++cnt > 100)
{
PrintLog(0,KERN_ERR "Error wait remap_count=%u\n",wait_count);
break;
}
msleep_interruptible(100);
}
if (rm_thread !=NULL)
{
wake_up_process(rm_thread);
kthread_stop(rm_thread);
}
if (rm_cache !=NULL)
{
cnt=atomic_read(&rm_cache_count);
kmem_cache_destroy(rm_cache);
if (cnt !=0) PrintLog(0,KERN_ERR "Error. Unfree %d lbb_remap items\n",cnt);
}
x_spin_lock_destroy(rm_bio_lock);
rm_bio_lock=NULL;
x_spin_lock_destroy(rm_it_lock);
rm_it_lock=NULL;
x_spin_lock_destroy(wait_lock);
wait_lock=NULL;
}
