#include <linux/kernel.h>
#include <linux/module.h>
#include "errors.h"
#include "lbb_kernel.h"
#include "lbb_device.h"
#include "lbb_device2.h"
#include "lbb_sys_fs.h"
#include "lbb_block_store.h"
#include "kernel_hook.h"
#include "lbb_remap.h"
#define FILE_NAME_ABSTRACT "addons"
#define FILE_NAME_MAP "bitmap.bin"
#define FILE_NAME_BLOCK "block.bin"
static struct kobject * lbb_root_kobj=NULL;
static const char * skip_map_file=FILE_NAME_MAP;
static const char * abstract_file=FILE_NAME_ABSTRACT;
static const char * block_map_file=FILE_NAME_BLOCK;
#pragma pack(push,1)
struct lbb_info
{
unsigned int version;
unsigned short info_size;
unsigned char status;
unsigned char store_mode;
unsigned short users;
unsigned short blk_size;
unsigned int blk_total;
unsigned int blk_busy;
unsigned short _free2_;
unsigned short _free3_;
unsigned long map_total;
unsigned long map_unused;
};
#pragma pack(pop)
struct lbb_dev_info
{
struct bin_attribute map_bin_attr;
struct bin_attribute abstract_attr;
struct bin_attribute table_bin_attr;
};
static ssize_t save_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
struct lbb_device * xdev=lbb_kobj_find_dev(kobj);
return lbb_xstore_cmd_info(xdev,buf);
}
static ssize_t orig_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
uint64 sect_beg,sect_end;
struct lbb_device * xdev=lbb_kobj_find_dev(kobj);
if (!xdev) return sprintf(buf,"?");
sect_beg=lbb_get_sect_beg(xdev);
sect_end=lbb_get_sect_end(xdev);
return sprintf(buf,"%s %llu %llu (%s)",lbb_get_orig_name(xdev),sect_beg,sect_end,format_mem_size((sect_end-sect_beg) << SECTOR_SHIFT));
}
static ssize_t info_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
int tmp;
struct lbb_info * info=(struct lbb_info *)buf;
struct lbb_device * xdev=lbb_kobj_find_dev(kobj);
if (!xdev) return 0;
info->info_size=sizeof(struct lbb_info);
info->version=get_module_version();
tmp=lbb_dev_get_status(xdev);
if ((tmp < 0)||(tmp > 3)) info->status=3;
else info->status=(unsigned char)tmp;
tmp=lbb_remap_get_users(xdev->rmp_dev);
if ((tmp < 0)||(tmp > 0xffff)) info->users=0xffff;
else info->users=(unsigned short)tmp;
tmp=lbb_block_store_status(xdev->store);
if ((tmp < 0)||(tmp > 3)) info->store_mode=3;
else info->store_mode=(unsigned char)tmp;
if (xdev->block_size > 0xffff) info->blk_size=0;
else info->blk_size=xdev->block_size;
if (xdev->store)
{
info->blk_busy=lbb_block_store_get_busy(xdev->store);
info->blk_total=lbb_block_store_get_total(xdev->store);
}
else
{
info->blk_busy=0;
info->blk_total=0;
}
if (xdev->skip)
{
info->map_total=lbb_map_tree_get_size(xdev->skip);
info->map_unused=lbb_map_tree_get_set(xdev->skip);
}
else
{
info->map_total=0;
info->map_unused=0;
}
return info->info_size;
}
static ssize_t stat_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
ssize_t rcod;
unsigned int users;
int status;
struct lbb_device * xdev=lbb_kobj_find_dev(kobj);
if (!xdev) return sprintf(buf,"?");
users=lbb_remap_get_users(xdev->rmp_dev);
status=lbb_dev_get_status(xdev);
switch (status)
{
default:
rcod=sprintf(buf,"Unknown status (%d)",status);
break;
case LBB_STAT_CATCH:
rcod=sprintf(buf,"CATCH");
break;
case LBB_STAT_OFF:
rcod=sprintf(buf,"OFF");
break;
case LBB_STAT_STOP:
rcod=sprintf(buf,"STOP (%s)",error_to_text(xdev->state_info));
break;
}
if (users !=0) rcod+=sprintf(buf+rcod," Users %u",users);
return rcod;
}
static ssize_t store_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
struct lbb_device * xdev=lbb_kobj_find_dev(kobj);
return lbb_xstore_print_info(xdev,buf);
}
static ssize_t bitmap_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
struct lbb_device * xdev=lbb_kobj_find_dev(kobj);
return lbb_xbitmap_print_info(xdev,buf);
}
ssize_t block_map_read(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
struct lbb_device * xdev=(struct lbb_device *)(attr->private);
return lbb_blockmap_read(xdev,data,offset,size);
}
ssize_t skip_map_read(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
struct lbb_device * xdev=(struct lbb_device *)(attr->private);
return lbb_bitmap_read(xdev,data,offset,size);
}
ssize_t skip_map_write(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
struct lbb_device * xdev=(struct lbb_device *)(attr->private);
return lbb_bitmap_write(xdev,data,offset,size);
}
ssize_t table_read(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
struct lbb_device * xdev=(struct lbb_device *)(attr->private);
return lbb_table_read(xdev,data,offset,size);
}
ssize_t table_write(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
return 0;
}
ssize_t abstract_read(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
ssize_t rcod;
struct lbb_device * xdev=(struct lbb_device *)(attr->private);
if (offset) return 0;
rcod=lbb_abstract_read(xdev,data,size);
return rcod;
}
ssize_t abstract_write(struct file * file,struct kobject * kobj,struct bin_attribute * attr,char * data,loff_t offset,size_t size)
{
ssize_t rcod;
struct lbb_device * xdev=(struct lbb_device *)(attr->private);
if (offset) return 0;
rcod=lbb_abstract_write(xdev,data,size);
return rcod;
}
static struct kobj_attribute lbb_attr_orig=__ATTR_RO(orig);
static struct kobj_attribute lbb_attr_save=__ATTR_RO(save);
static struct kobj_attribute lbb_attr_stat=__ATTR_RO(stat);
static struct kobj_attribute lbb_attr_store=__ATTR_RO(store);
static struct kobj_attribute lbb_attr_bitmap=__ATTR_RO(bitmap);
static struct kobj_attribute lbb_attr_info=__ATTR_RO(info);
int lbb_sysfs_disk_add(struct lbb_device * xdev)
{
ssize_t size;
int rcod;
const char * name;
struct lbb_dev_info * info;
if (xdev==NULL) return ERR_ZERO_POINTER;
if ((xdev->info !=NULL)||
(xdev->kobj !=NULL)) return ERR_INTERNAL5;
info=xkmalloc(sizeof(struct lbb_dev_info));
if (info==NULL) return ERR_NO_MEMORY;
xmemset(info,0,sizeof(struct lbb_dev_info));
name=lbb_remap_disk_name(xdev->rmp_dev);
xdev->kobj=kobject_create_and_add(name,lbb_root_kobj);
if (!xdev->kobj)
{
PrintLog(0,KERN_ERR "failed to create /sys/root/../%s",name);
return ERR_KOBJ_SYS;
}
{
info->map_bin_attr.attr.name=skip_map_file;
info->map_bin_attr.attr.mode=0666;
info->map_bin_attr.private=xdev;
info->map_bin_attr.size=(xdev->block_total+7) / 8;
info->map_bin_attr.read=skip_map_read;
info->map_bin_attr.write=skip_map_write;
rcod=sysfs_create_bin_file(xdev->kobj,&(info->map_bin_attr));
}
{
info->table_bin_attr.attr.name=block_map_file;
info->table_bin_attr.attr.mode=0666;
info->table_bin_attr.private=xdev;
info->table_bin_attr.size=(xdev->block_total+7) / 8;
info->table_bin_attr.read=block_map_read;
info->table_bin_attr.write=NULL;
rcod=sysfs_create_bin_file(xdev->kobj,&(info->table_bin_attr));
}
size=4096;
{
info->abstract_attr.attr.name=abstract_file;
info->abstract_attr.attr.mode=0666;
info->abstract_attr.private=xdev;
info->abstract_attr.size=size;
info->abstract_attr.read=abstract_read;
info->abstract_attr.write=abstract_write;
rcod=sysfs_create_bin_file(xdev->kobj,&(info->abstract_attr));
}
rcod=sysfs_create_file(xdev->kobj,&(lbb_attr_save.attr));
rcod=sysfs_create_file(xdev->kobj,&(lbb_attr_orig.attr));
rcod=sysfs_create_file(xdev->kobj,&(lbb_attr_stat.attr));
rcod=sysfs_create_file(xdev->kobj,&(lbb_attr_store.attr));
rcod=sysfs_create_file(xdev->kobj,&(lbb_attr_bitmap.attr));
rcod=sysfs_create_file(xdev->kobj,&(lbb_attr_info.attr));
xdev->info=info;
return NO_ERR;
}
void lbb_sysfs_disk_del(struct lbb_device * xdev)
{
if (xdev !=NULL)
{
if (xdev->kobj) kobject_put(xdev->kobj);
if (xdev->info) xkfree(xdev->info);
xdev->kobj=NULL;
xdev->info=NULL;
}
}
static ssize_t cmd_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
return get_module_info(buf);
}
static ssize_t cmd_store(struct kobject * kobj,struct kobj_attribute * attr,const char * buf,size_t count)
{
return lbb_cmd_parser(buf,count);
}
static ssize_t debug_show(struct kobject * kobj,struct kobj_attribute * attr,char * buf)
{
return get_debug_info(buf);
}
static struct kobj_attribute lbb_attr_cmd=__ATTR(cmd,0444,cmd_show,cmd_store);
static struct kobj_attribute lbb_attr_debug=__ATTR(debug,0444,debug_show,NULL);
int lbb_sysfs_init(const char * name)
{
int rcod;
lbb_root_kobj=kobject_create_and_add(name,kernel_kobj);
if (!lbb_root_kobj)
{
PrintLog(0,KERN_ERR "failed to create root kobject\n");
return ERR_KOBJ_CREATE;
}
lbb_attr_cmd.attr.mode=00666;
rcod=sysfs_create_file(lbb_root_kobj,&(lbb_attr_cmd.attr));
if (rcod)
{
PrintLog(0,KERN_ERR "failed to create file </sys/kernel/../cmd>\n");
return ERR_KOBJ_FILE;
}
rcod=sysfs_create_file(lbb_root_kobj,&(lbb_attr_debug.attr));
if (rcod)
{
PrintLog(0,KERN_ERR "failed to create file </sys/kernel/../debug>\n");
return ERR_KOBJ_FILE;
}
return NO_ERR;
}
void lbb_sysfs_remove(void)
{
kobject_put(lbb_root_kobj);
}
